#
#  - Midi sysex procedures -
#
#   Created:    07/00
#   Changed:    03/01 
#
########################################################################

################################################################
#
# bin2int --
#
# 8-bit binary digits to unsigned integer conversion
#
# Arguments:     digits ... binary digits like "00001111"
#
# Returns:    unsigned integer value of the binary Digit
#
proc bin2int {digits} {
    set binary_value [binary format B8 $digits]
    binary scan $binary_value c signed_value
    set unsigned_value [expr ( $signed_value + 0x100 ) % 0x100]
    return $unsigned_value
}   

################################################################
#
# GetEntryNumber --
#
# Gets the address entry number inside a patch (0..F)
#
# Arguments:     line ... a line from sysex file
#
# Returns:    nr 
#
proc GetEntryNumber {line} {
    binary scan $line H12cccH* syx_begin addr1 addr2 addr3 syx_end
	set nr [expr $addr3 & 15] 
#	puts -nonewline [format "%0.2x" $addr2][format "%0.2x" $addr3]
    return $nr

}   

################################################################
#
# GetBaseAddress --
#
# Gets the base address of the patch inside the map 
#
# Arguments:     line ... a line from sysex file
#
# Returns:    address in int format 
#
proc GetBaseAddress {line} {
    binary scan $line H12cccH* syx_begin addr1 addr2 addr3 syx_end
#	puts [format "%0.2x" $addr1][format "%0.2x" $addr2][format "%0.2x" $addr3]
	set address [expr $addr2 * 256 + $addr3]
    return $address

}   

################################################################
#
# GetPatch --
#
# Gets the patch number from address field
#
# Arguments:     line ... a line from sysex file
#
# Returns:    patch number 
#
proc GetPatch {line} {
    binary scan $line H12cccH* syx_begin addr1 addr2 addr3 syx_end
    set address [expr $addr2 * 128 + $addr3]
    set patch [expr int( $address / 16)]
	
	# assign bank (0...A, 1...B)
	set bankval [expr $patch / 16]
	if { $bankval == 0 } {
		set bank "A"
	} else {
		set bank "B"
		incr patch -16
	}
	incr patch	
	
    return "$bank-$patch"

}   

################################################################
#
# GetDataValue --
#
# Gets the data value from the nibbled values
#
# Arguments:     line ... a line from sysex file
#
# Returns:    value
#
proc GetDataValue {line} {
    binary scan $line H18ccH* syx_begin int1 int2 syx_end
    set value [expr $int1 * 16 + $int2]
#	puts -nonewline " [format "%0.2x" $int1]:[format "%0.2x" $int2] "
    return $value
}   

################################################################
#
# GetChecksumValue --
#
# Gets the checksum value from file
#
# Arguments:     line ... a line from sysex file
#
# Returns:    value
#
proc GetChecksumValue {line} {
    binary scan $line H22cH* syx_begin checksum syx_end
#	puts  " [format "%0.2x" $checksum]"
    return $checksum
} 
  
################################################################
#
# SetChecksum --
#
# Sets the checksum value, Attention! clearing bit 7 
#
# Arguments:   addr1 addr2 data1 data22 ... hex strings 
#
# Returns:    checksum
#
proc SetChecksum {addr1 addr2 data2 data3} {
    set sum [expr $addr1 + $addr2 + $data2 + $data3]
	set remainder [expr $sum % 128]
	set checksum [expr (128 - $remainder) & 127]
    return $checksum
}   

################################################################
#
# Unmask --
#
# Extract the data from the data value by magic masking and shifting
#
# Arguments:     data ... data value
#        mask ... binary digit mask
#        lshift ... left shift value
#
# Returns:    value
#
proc Unmask {data {mask "00000000"} {lshift 0}} {
    set dmask [bin2int $mask]
    set value [expr ( $data & $dmask ) >> $lshift]
    return $value
}   

################################################################
#
# ReceiveMidi --
#
# Record sysex data from MINI Input device or file. 
#
# Arguments:  	patchname ... patch name ("A-1" ... "B-16") or "0-0" for all
#
# Returns:    -
#
# Note: Patch 0-0 provides whole data to record
#
proc ReceiveMidi {patchname} {
global map recmap def a33

	set a33(cancel) 0

	set waitwin [DisplayWaitForInput] 

	UpdateDisplay "WAITING ..." "PLEASE DUMP MIDI!"

	set f [midi::openin $def(midiin)]
#	set f [open dump.syx]

	fconfigure $f -translation binary

	set len 0
	while { 1 } {

		# check for user break
		if {$a33(cancel)} {
			UpdateDisplay "PROCESS CANCELED" "BY USER"
			close $f
			return
		}

		# read one sysex command line 
	    	set line [read $f 13]

	    	if [string length $line] {
    
			# destroy wait message window
			if {[winfo exists $waitwin]} {
				destroy $waitwin
			} 

			incr len 13		
			set patch [GetPatch $line]
			set nr [GetEntryNumber $line]
        		set value [GetDataValue $line]
#			set c [GetChecksumValue $line]
			
			UpdateDisplay "PROCESSING $patch "

			set recmap($patch,name) "{}"
			set recmap($patch,description) "{}"
			set recmap($patch,lower_name) "{}"
			set recmap($patch,upper_name) "{} "

			## debugging			
			# eval binary scan "\\x[format "%x" $value]" B* bin
			# puts "$patch - $bin"
		
			switch $nr {
		
				0 	{
						# Initialize recmap address field
						set recmap($patch,address) [GetBaseAddress $line]
				
						# Velocity sensitivy, Bit 1-0, Value: H L M
						set recmap($patch,velocity) [Unmask $value 00000011 0]
						
						# Transpose On/Off, Bit 2
						set recmap($patch,isTransposed) [Unmask $value 00000100 2]
						
						# Transpose Value, Bit 7-3, Value -5/+6
						set recmap($patch,transpose) [Unmask $value 11111000 3]
					}
		
				1 	{
						# Tempo in BPM, Bit 7-0, Value: 20-250
						set recmap($patch,tempo) [Unmask $value 11111111 0]
					}
					
				2 	{
						# Lower Octave, Bit 2-0, Value: -2,+2
						set recmap($patch,lower_octave) [Unmask $value 00000111 0]

						# Upper Octave, Bit 5-3, Value: -2,+2
						set recmap($patch,upper_octave) [Unmask $value 00111000 3]

						# Dummy bit, Bit 6
						set recmap($patch,dummy) [Unmask $value 01000000 6]

						# Clock, Bit 7, Value: 0=internal, 1=external
						set recmap($patch,clock) [Unmask $value 10000000 7]
					}
					
					
				3 	{
						# Split Value, Bit 5-0, Value: 0-60, where 0 is the first B ( 8th key from left)
						set recmap($patch,split) [Unmask $value 00111111 0]

						# Split On/Off, Bit 6
						set recmap($patch,isSplitted) [Unmask $value 01000000 6]

						# Lever Assign, Bit 7, Value: 0=Modulation, 1=Aftertouch
						set recmap($patch,lever) [Unmask $value 10000000 7]
					}
					
				4 	{
						# Data Entry Assign, Bit 7-0, Value: 0-255 (Controllers)
						set recmap($patch,data_entry) [Unmask $value 11111111 0]
					}

				5 	{
						# Upper CC00 Value, Bit 6-0, Value: 0-127
						set recmap($patch,upper_cc00) [Unmask $value 01111111 0]

						# Upper Sustain Tx On/Off, Bit 7
						set recmap($patch,upper_isSustain) [Unmask $value 10000000 7]
					}
					
				6 	{
						# Upper CC32 Value, Bit 6-0, Value: 0-127
						set recmap($patch,upper_cc32) [Unmask $value 01111111 0]

						# Upper PitchBend & Lever Tx On/Off, Bit 7
						set recmap($patch,upper_isPBLever) [Unmask $value 10000000 7]
					}

				7 	{
						# Upper Program Change Value, Bit 6-0, Value: 0-127
						set recmap($patch,upper_program) [Unmask $value 01111111 0]

						# Upper Reverb On/Off, Bit 7
						set recmap($patch,upper_isReverb) [Unmask $value 10000000 7]
					}
					
				8 	{
						# Upper Tx Midid Channel, Bit 3-0, Value: 0-15
						set recmap($patch,upper_midichannel) [Unmask $value 00001111 0]

						# Lower Keyboard Mode On/Off, Bit 4
						set recmap($patch,isLower) [Unmask $value 00010000 4]

						# Upper Keyboard Mode On/Off, Bit 5
						set recmap($patch,isUpper) [Unmask $value 00100000 5]

						# Lower Expression Pedal Tx On/Off, Bit 6
						set recmap($patch,lower_isExpression) [Unmask $value 01000000 6]

						# Upper Expression Pedal Tx On/Off, Bit 7
						set recmap($patch,upper_isExpression) [Unmask $value 10000000 7]
					}
					
				9 	{
						# Lower CC00 Value, Bit 6-0, Value: 0-127
						set recmap($patch,lower_cc00) [Unmask $value 01111111 0]

						# Lower Sustain Tx On/Off, Bit 7
						set recmap($patch,lower_isSustain) [Unmask $value 10000000 7]
					}
					
			     10 	{
						# Lower CC32 Value, Bit 6-0, Value: 0-127
						set recmap($patch,lower_cc32) [Unmask $value 01111111 0]

						# Lower PitchBend & Lever Tx On/Off, Bit 7
						set recmap($patch,lower_isPBLever) [Unmask $value 10000000 7]
					}

				11 	{
						# Lower Program Change Value, Bit 6-0, Value: 0-127
						set recmap($patch,lower_program) [Unmask $value 01111111 0]

						# Lower Reverb On/Off, Bit 7
						set recmap($patch,lower_isReverb) [Unmask $value 10000000 7]
					}

				12 	{
						# Lower Tx Midid Channel, Bit 3-0, Value: 0-15
						set recmap($patch,lower_midichannel) [Unmask $value 00001111 0]

						# Lower Chorus On/Off, Bit 4
						set recmap($patch,lower_isChorus) [Unmask $value 00010000 4]

						# Upper Chorus On/Off, Bit 5
						set recmap($patch,upper_isChorus) [Unmask $value 00100000 5]

						# Midi Out 1 On/Off, Bit 6
						set recmap($patch,isMidiOut2) [Unmask $value 01000000 6]

						# Midid Out 2 On/Off, Bit 7
						set recmap($patch,isMidiOut1) [Unmask $value 10000000 7]
					}
					
			      13 	{
				
						# Upper Volume, Bit 3-0, Value: 0-127
						set recmap($patch,upper_volume) [Unmask $value 00001111 0]

						# Lower Volume, Bit 7-4, Value: 0-127
						set recmap($patch,lower_volume) [Unmask $value 11110000 4]
					}
					
			      14 	{
						# Upper Reverb, Bit 3-0, Value: 0-127
						set recmap($patch,upper_reverb) [Unmask $value 00001111 0]

						# Lower Reverb, Bit 7-4, Value: 0-127
						set recmap($patch,lower_reverb) [Unmask $value 11110000 4]
					}
					
			      15 	{
						# Upper Chorus, Bit 3-0, Value: 0-127
						set recmap($patch,upper_chorus) [Unmask $value 00001111 0]

						# Lower Chorus, Bit 7-4, Value: 0-127
						set recmap($patch,lower_chorus) [Unmask $value 11110000 4]

						# Finish after last patch or current patch reached 
						if { ($patch == "B-16") || ($patch == $patchname) } {
							break
						}

					}
			}
			
    		} 

		update
	}

	close $f
	UpdateDisplay "$len BYTES READ" OK

	set diff [DiffMap $patchname]
	if { $diff > 1 } {
		
		incr diff -1
		if { $patchname == "0-0" } {
			set msg "Received patches differs in $diff cases from the current patches!\nWould you like to overwrite ALL current patches?"
		} else {
			set msg "Received patch '$patchname' differs in $diff cases from the current patch!\nWould you like to overwrite the current patch?"
		}
		set rc [tk_messageBox -title Diff -message $msg -type yesno]
		if { $rc == "no" } {return}
	} else {
		set msg "Patches are identical!"
		set rc [tk_messageBox -title Diff -message $msg -type ok]
		return
	}	

	# Overwrite patch maps
	if { $patchname == "0-0" } {

		#  overwrites the whole map
		foreach name [array name recmap] {
			set map($name) $recmap($name)
		}

		for {set i 1} {$i <= 16} {incr i} {
			SaveMap "A-$i"
			SaveMap "B-$i"
			eval .view.workspace.bp itemconfigure "A-$i" -text $map(A-$i,name)
			eval .view.workspace.bp itemconfigure "B-$i" -text $map(B-$i,name)
		}
		LoadPatch

	} else {
			
		# overwrites only the current map
		foreach name [array name recmap $patchname,*] {
			set map($name) $recmap($name)
		}
		SaveMap $patchname
		LoadPatch
	}



}

################################################################
#
# SendMidi --
#
# Writes sysex data to MIDI OUT
#
# Arguments:  	patchname ... patch name ("A-1" ... "B-16") or "0-0" for all
#
# Returns:    -
#
# Note: Patch 0-0 provides whole data to write 
#
proc SendMidi {patchname} {
global a33 map def

	if { $patchname == "0-0" } {
		set patch_list {}
		for {set i 1} {$i <= 16} {incr i} {
			lappend patch_list "A-$i"
		}
		for {set i 1} {$i <= 16} {incr i} {
			lappend patch_list "B-$i"
		}
	} else {
		set patch_list $patchname
	}

	set a33(cancel) 0
	UpdateDisplay "PROCESSING ..."
	set f [midi::openout $def(midiout)]
	fconfigure $f -translation binary
	foreach patch $patch_list {

		LoadMap $patch

		for {set nr 0} {$nr <= 15} {incr nr} {

			set address [expr $map($patch,address) + $nr]
			set addr1 "00"
			set addr2 [format "%0.2x" [expr $address >> 8]]
			set addr3 [format "%0.2x" [expr $address & 255]]
			
			set value 0
			switch $nr {
		
				0 	{
						incr value $map($patch,velocity)
						incr value [expr $map($patch,isTransposed) << 2]
						incr value [expr $map($patch,transpose) << 3]
					}

				1 	{
						incr value $map($patch,tempo)
					}
					
				2 	{
						incr value $map($patch,lower_octave)
						incr value [expr $map($patch,upper_octave) << 3]
						incr value [expr $map($patch,dummy) << 6]
						incr value [expr $map($patch,clock) << 7]
					}
					
				3 	{
						incr value $map($patch,split)
						incr value [expr $map($patch,isSplitted) << 6]
						incr value [expr $map($patch,lever) << 7]
					}
					
				4 	{
						incr value $map($patch,data_entry)
					}

				5 	{
						incr value $map($patch,upper_cc00)
						incr value [expr $map($patch,upper_isSustain) << 7]
					}
					
				6 	{
						incr value $map($patch,upper_cc32)
						incr value [expr $map($patch,upper_isPBLever) << 7]
					}

				7 	{
						incr value $map($patch,upper_program)
						incr value [expr $map($patch,upper_isReverb) << 7]
					}
					
				8 	{
						incr value $map($patch,upper_midichannel)
						incr value [expr $map($patch,isLower) << 4]
						incr value [expr $map($patch,isUpper) << 5]
						incr value [expr $map($patch,lower_isExpression) << 6]
						incr value [expr $map($patch,upper_isExpression) << 7]
					}
					
				9 	{
						incr value $map($patch,lower_cc00)
						incr value [expr $map($patch,lower_isSustain) << 7]
					}
					
			      10 	{
						incr value $map($patch,lower_cc32)
						incr value [expr $map($patch,lower_isPBLever) << 7]
					}

				11 	{
						incr value $map($patch,lower_program)
						incr value [expr $map($patch,lower_isReverb) << 7]
					}

				12 	{
						incr value $map($patch,lower_midichannel)
						incr value [expr $map($patch,lower_isChorus) << 4]
						incr value [expr $map($patch,upper_isChorus) << 5]
						incr value [expr $map($patch,isMidiOut2) << 6]
						incr value [expr $map($patch,isMidiOut1) << 7]
					}
					
			      13 	{
						incr value $map($patch,upper_volume)
						incr value [expr $map($patch,lower_volume) << 4]
					}
					
			      14 	{
						incr value $map($patch,upper_reverb)
						incr value [expr $map($patch,lower_reverb) << 4]
					}
					
			      15 	{
						incr value $map($patch,upper_chorus)
						incr value [expr $map($patch,lower_chorus) << 4]
					}

			}
			set addr1 [expr $address >> 8]
			set addr2 [expr $address & 255]
			set data1 [expr $value >> 4]
			set data2 [expr $value & 15]
			set checksum [SetChecksum $addr1 $addr2 $data1 $data2]

			# write sysex data to file
			puts -nonewline $f "\xf0\x41\x10\x00\x02\x12"		;# sysex start
			puts -nonewline $f "\x00"					;# addr0
			eval puts -nonewline $f "\\x[format "%x" $addr1]"	;# addr1
			eval puts -nonewline $f "\\x[format "%x" $addr2]"	;# addr2
			eval puts -nonewline $f "\\x[format "%x" $data1]"	;# data1
			eval puts -nonewline $f "\\x[format "%x" $data2]"	;# data2
			eval puts -nonewline $f "\\x[format "%x" $checksum]" 	;# checksum
			puts -nonewline $f "\xf7"					;# sysex end
			flush $f
			after 50
			if {$a33(cancel)} {
				UpdateDisplay "PROCESS CANCELED" "BY USER"
				close $f
				return
			}
			UpdateDisplay "PROCESSING $patch ..." [string repeat "=" $nr]

			#puts "[format %0.4x $address] [format %0.2x $data1]:[format %0.2x $data2] [format %0.2x $checksum]"

		}
		

	}

	close $f
	UpdateDisplay "MIDI DATA SENT" OK
}

################################################################
#
# CreateFile --
#
# Open file Dialog. Select dump file name. Create configuration
# files for all patches from dump file
#
# Arguments:   -
#
# Returns:    filename ... dump file
#
proc CreateFiles {} {
global map

 	set types {{"sysex file" {.syx}}}
	set filename [tk_getOpenFile -filetypes $types -parent .]

	if { $filename == "" } {
		return ""
	}
	
	
	ReadDataFile $filename
	for {set i 1} {$i <= 16} {incr i} {
		SaveMap "A-$i"
	}
	for {set i 1} {$i <= 16} {incr i} {
		SaveMap "B-$i"
	}

	return $filename
}


################################################################
#
# LoadDefFile --
#
# Load misc options from file
#
# Arguments:  -
#
# Returns:    -
#
proc LoadDefFile {} {
global def projdir

	set filename [file join $projdir bin a33.ini]
	if {![file exist $filename]} {
		return
	}

	set f [open $filename r]

	while {![eof $f]} {
		gets $f s
		eval $s
	}

	close $f
}

################################################################
#
# SaveDefFile --
#
# Save Load misc options into file
#
# Arguments:  -
#
# Returns:    -
#
proc SaveDefFile {} {
global def projdir

	set filename [file join $projdir bin a33.ini]

	set f [open $filename w]
	
	puts $f "set def(midiin) $def(midiin)"
	puts $f "set def(midiout) $def(midiout)"


	close $f
}

################################################################
#
# PrintToFile --
#
# Prints the patch labels to the selected file name
#
# Arguments:  -
#
# Returns:    -
#
proc PrintToFile {} {
global map projdir

	set filename [file join $projdir labels a33_template.rtf]

	if {![file exist $filename]} {
		set msg "Fatal Error!\n The specified Template $filename doesn't exist!"
		set rc [tk_messageBox -title Error -message $msg -type ok]
		return
	}

	# read template file
	set f [open $filename r]
	set s [read $f]
	close $f
	
	# replace patch names 1-16
	for {set i 1} {$i <= 16} {incr i} {
		regsub "A$i." $s "$map(A-$i,name)" s
		regsub "B$i." $s "$map(B-$i,name)" s
	}

	# save label file
	set filename [tk_getSaveFile -filetypes {{"Rich Text Format" rtf}} -parent . -initialfile a33label -defaultextension .rtf]

	if { $filename == "" } {
		return
	}
	
	set f [open $filename w]
	puts -nonewline $f $s
	close $f

}

#############################################################################
#
# DisplayWaitForInput --
#
# Display message while waiting for sysex data from MIDI input
# contains cancel button to escape from action
#
# Arguments:  -
#
# Returns:    w ... message window
#
#	sets a33(cancel) 
#
proc DisplayWaitForInput {} {
global a33

	# new window
	set w .msg

	set header "Ready to receive MIDI data bulk dump!"
	set msg "Please press \nEDIT, MIDI/PARAM, 2, 8, then ENTER \non your A33 ..."

	catch {destroy $w}
 	toplevel $w -bd 0
	wm transient $w .
	wm title $w "Receiving..."
      wm iconname $w "Receiving..."
	wm transient $w .
	grab set -global $w

	frame $w.msg -bd 1 -relief flat
	  pack $w.msg -in $w -side top -expand 1 -fill both 

	frame $w.control -bd 1 -relief flat
	  pack $w.control -in $w -side bottom -fill x
	button $w.control.cancel -width 12 -text "Cancel" -under 0 -command "set a33(cancel) 1; grab release $w; destroy $w"
	  pack $w.control.cancel -in $w.control -side left -expand 1 -padx 3m -pady 2m

	label $w.msg.header -width 40 -text $header -anchor c
	  pack $w.msg.header -in $w.msg -side top -fill x -padx 10 -pady 10 
	label $w.msg.label -text $msg -anchor c 
	  pack $w.msg.label -in $w.msg -side top -fill x -padx 10 -pady 10 

	wm withdraw $w
	update idletasks
	set x [expr [winfo x .] + ([winfo width .] / 2) - ([winfo reqwidth $w] / 2)]
	set y [expr [winfo y .] + ([winfo height .] / 2) - ([winfo reqheight $w] / 2)]
	wm geometry $w +$x+$y
	update idletasks
	wm deiconify $w
	update idletasks

	return $w
}

